from __future__ import annotations

import asyncio
import logging
import sys
from typing import Any, Dict, Optional, Type

import backoff
import httpx

from apitally.client.base import (
    ApitallyClientBase,
    ApitallyKeyCacheBase,
    handle_retry_giveup,
)


logger = logging.getLogger(__name__)
retry = backoff.on_exception(
    backoff.expo,
    httpx.HTTPError,
    max_tries=3,
    on_giveup=handle_retry_giveup,
    raise_on_giveup=False,
)


class ApitallyClient(ApitallyClientBase):
    def __init__(
        self,
        client_id: str,
        env: str,
        sync_api_keys: bool = False,
        sync_interval: float = 60,
        key_cache_class: Optional[Type[ApitallyKeyCacheBase]] = None,
    ) -> None:
        super().__init__(
            client_id=client_id,
            env=env,
            sync_api_keys=sync_api_keys,
            sync_interval=sync_interval,
            key_cache_class=key_cache_class,
        )
        self._stop_sync_loop = False
        self._sync_loop_task: Optional[asyncio.Task[Any]] = None

    def get_http_client(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(base_url=self.hub_url, timeout=1)

    def start_sync_loop(self) -> None:
        self._stop_sync_loop = False
        self._sync_loop_task = asyncio.create_task(self._run_sync_loop())

    async def _run_sync_loop(self) -> None:
        if self.sync_api_keys:
            try:
                async with self.get_http_client() as client:
                    await self.get_keys(client)
            except Exception as e:
                logger.exception(e)
        while not self._stop_sync_loop:
            try:
                await asyncio.sleep(self.sync_interval)
                async with self.get_http_client() as client:
                    await self.send_requests_data(client)
                    if self.sync_api_keys:
                        await self.get_keys(client)
            except Exception as e:  # pragma: no cover
                logger.exception(e)

    def stop_sync_loop(self) -> None:
        self._stop_sync_loop = True

    def send_app_info(self, app_info: Dict[str, Any]) -> None:
        payload = self.get_info_payload(app_info)
        asyncio.create_task(self._send_app_info(payload=payload))

    async def send_requests_data(self, client: httpx.AsyncClient) -> None:
        payload = self.get_requests_payload()
        await self._send_requests_data(client, payload)

    async def get_keys(self, client: httpx.AsyncClient) -> None:
        if response_data := await self._get_keys(client):  # Response data can be None if backoff gives up
            self.handle_keys_response(response_data)
        elif self.key_registry.salt is None:
            logger.error("Initial Apitally API key sync failed")
            # Exit because the application will not be able to authenticate requests
            sys.exit(1)

    @retry
    async def _send_app_info(self, payload: Dict[str, Any]) -> None:
        async with self.get_http_client() as client:
            response = await client.post(url="/info", json=payload, timeout=1)
            if response.status_code == 404 and "Client ID" in response.text:
                self.stop_sync_loop()
                logger.error(f"Invalid Apitally client ID {self.client_id}")
            else:
                response.raise_for_status()

    @retry
    async def _send_requests_data(self, client: httpx.AsyncClient, payload: Dict[str, Any]) -> None:
        response = await client.post(url="/requests", json=payload)
        response.raise_for_status()

    @retry
    async def _get_keys(self, client: httpx.AsyncClient) -> Dict[str, Any]:
        response = await client.get(url="/keys")
        response.raise_for_status()
        return response.json()



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, List, Optional

from ninja.security import APIKeyHeader

from apitally.client.asyncio import ApitallyClient
from apitally.client.base import KeyInfo
from apitally.django import ApitallyMiddleware as _ApitallyMiddleware
from apitally.django import DjangoViewInfo


if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse
    from ninja import NinjaAPI


__all__ = ["ApitallyMiddleware", "APIKeyAuth", "APIKeyAuthBase", "KeyInfo"]


class ApitallyMiddleware(_ApitallyMiddleware):
    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]) -> None:
        super().__init__(get_response)
        if self.client.sync_api_keys:
            api = _get_api(self.views)
            _add_exception_handlers(api)


class AuthError(Exception):
    pass


class InvalidAPIKey(AuthError):
    pass


class PermissionDenied(AuthError):
    pass


class APIKeyAuthBase(APIKeyHeader):
    def __init__(self, scopes: Optional[List[str]] = None) -> None:
        self.scopes = scopes or []

    def authenticate(self, request: HttpRequest, key: Optional[str]) -> Optional[KeyInfo]:
        if key is None:
            return None
        if self.param_name == "Authorization":
            scheme, _, api_key = key.partition(" ")
            if scheme.lower() != "apikey":
                return None
        else:
            api_key = key
        key_info = ApitallyClient.get_instance().key_registry.get(api_key)
        if key_info is None:
            raise InvalidAPIKey()
        if not key_info.has_scopes(self.scopes):
            raise PermissionDenied()
        return key_info


class APIKeyAuth(APIKeyAuthBase):
    param_name = "Authorization"
    openapi_description = "Provide your API key using the <code>Authorization</code> header and the scheme prefix <code>ApiKey</code>.<br>Example: <pre>Authorization: ApiKey your_api_key_here</pre>"


def _get_api(views: List[DjangoViewInfo]) -> NinjaAPI:
    try:
        return next(
            (view.func.__self__.api for view in views if view.is_ninja_path_view and hasattr(view.func, "__self__"))
        )
    except StopIteration:  # pragma: no cover
        raise RuntimeError("Could not find NinjaAPI instance")


def _add_exception_handlers(api: NinjaAPI) -> None:
    def on_invalid_api_key(request: HttpRequest, exc) -> HttpResponse:
        return api.create_response(request, {"detail": "Invalid API key"}, status=403)

    def on_permission_denied(request: HttpRequest, exc) -> HttpResponse:
        return api.create_response(request, {"detail": "Permission denied"}, status=403)

    api.add_exception_handler(InvalidAPIKey, on_invalid_api_key)
    api.add_exception_handler(PermissionDenied, on_permission_denied)



from __future__ import annotations

import json
import logging
import os
import re
import threading
from abc import ABC, abstractmethod
from collections import Counter
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from hashlib import scrypt
from math import floor
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union, cast
from uuid import UUID, uuid4


logger = logging.getLogger(__name__)

HUB_BASE_URL = os.getenv("APITALLY_HUB_BASE_URL") or "https://hub.apitally.io"
HUB_VERSION = "v1"

TApitallyClient = TypeVar("TApitallyClient", bound="ApitallyClientBase")


def handle_retry_giveup(details) -> None:  # pragma: no cover
    logger.exception("Apitally client failed to sync with hub: {target.__name__}(): {exception}".format(**details))


class ApitallyClientBase:
    _instance: Optional[ApitallyClientBase] = None
    _lock = threading.Lock()

    def __new__(cls, *args, **kwargs) -> ApitallyClientBase:
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(
        self,
        client_id: str,
        env: str,
        sync_api_keys: bool = False,
        sync_interval: float = 60,
        key_cache_class: Optional[Type[ApitallyKeyCacheBase]] = None,
    ) -> None:
        if hasattr(self, "client_id"):
            raise RuntimeError("Apitally client is already initialized")  # pragma: no cover
        try:
            UUID(client_id)
        except ValueError:
            raise ValueError(f"invalid client_id '{client_id}' (expected hexadecimal UUID format)")
        if re.match(r"^[\w-]{1,32}$", env) is None:
            raise ValueError(f"invalid env '{env}' (expected 1-32 alphanumeric lowercase characters and hyphens only)")

        self.client_id = client_id
        self.env = env
        self.sync_api_keys = sync_api_keys
        self.sync_interval = sync_interval
        self.instance_uuid = str(uuid4())
        self.request_logger = RequestLogger()
        self.validation_error_logger = ValidationErrorLogger()
        self.key_registry = KeyRegistry()
        self.key_cache = key_cache_class(client_id=client_id, env=env) if key_cache_class is not None else None

        if self.key_cache is not None and (key_data := self.key_cache.retrieve()):
            try:
                self.handle_keys_response(json.loads(key_data), cache=False)
            except (json.JSONDecodeError, TypeError, KeyError):  # pragma: no cover
                logger.exception("Failed to load keys from cache")

    @classmethod
    def get_instance(cls: Type[TApitallyClient]) -> TApitallyClient:
        if cls._instance is None:
            raise RuntimeError("Apitally client not initialized")  # pragma: no cover
        return cast(TApitallyClient, cls._instance)

    @property
    def hub_url(self) -> str:
        return f"{HUB_BASE_URL}/{HUB_VERSION}/{self.client_id}/{self.env}"

    def get_info_payload(self, app_info: Dict[str, Any]) -> Dict[str, Any]:
        payload = {
            "instance_uuid": self.instance_uuid,
            "message_uuid": str(uuid4()),
        }
        payload.update(app_info)
        return payload

    def get_requests_payload(self) -> Dict[str, Any]:
        requests = self.request_logger.get_and_reset_requests()
        validation_errors = self.validation_error_logger.get_and_reset_validation_errors()
        api_key_usage = self.key_registry.get_and_reset_usage_counts() if self.sync_api_keys else {}
        return {
            "instance_uuid": self.instance_uuid,
            "message_uuid": str(uuid4()),
            "requests": requests,
            "validation_errors": validation_errors,
            "api_key_usage": api_key_usage,
        }

    def handle_keys_response(self, response_data: Dict[str, Any], cache: bool = True) -> None:
        self.key_registry.salt = response_data["salt"]
        self.key_registry.update(response_data["keys"])

        if cache and self.key_cache is not None:
            self.key_cache.store(json.dumps(response_data, check_circular=False, allow_nan=False))


class ApitallyKeyCacheBase(ABC):
    def __init__(self, client_id: str, env: str) -> None:
        self.client_id = client_id
        self.env = env

    @property
    def cache_key(self) -> str:
        return f"apitally:keys:{self.client_id}:{self.env}"

    @abstractmethod
    def store(self, data: str) -> None:
        """Store the key data in cache as a JSON string."""
        pass  # pragma: no cover

    @abstractmethod
    def retrieve(self) -> str | bytes | bytearray | None:
        """Retrieve the stored key data from the cache as a JSON string."""
        pass  # pragma: no cover


@dataclass(frozen=True)
class RequestInfo:
    consumer: Optional[str]
    method: str
    path: str
    status_code: int


class RequestLogger:
    def __init__(self) -> None:
        self.request_counts: Counter[RequestInfo] = Counter()
        self.response_times: Dict[RequestInfo, Counter[int]] = {}
        self._lock = threading.Lock()

    def log_request(
        self, consumer: Optional[str], method: str, path: str, status_code: int, response_time: float
    ) -> None:
        request_info = RequestInfo(
            consumer=consumer,
            method=method.upper(),
            path=path,
            status_code=status_code,
        )
        response_time_ms_bin = int(floor(response_time / 0.01) * 10)  # In ms, rounded down to nearest 10ms
        with self._lock:
            self.request_counts[request_info] += 1
            self.response_times.setdefault(request_info, Counter())[response_time_ms_bin] += 1

    def get_and_reset_requests(self) -> List[Dict[str, Any]]:
        data: List[Dict[str, Any]] = []
        with self._lock:
            for request_info, count in self.request_counts.items():
                data.append(
                    {
                        "consumer": request_info.consumer,
                        "method": request_info.method,
                        "path": request_info.path,
                        "status_code": request_info.status_code,
                        "request_count": count,
                        "response_times": self.response_times.get(request_info) or Counter(),
                    }
                )
            self.request_counts.clear()
            self.response_times.clear()
        return data


@dataclass(frozen=True)
class ValidationError:
    consumer: Optional[str]
    method: str
    path: str
    loc: Tuple[str, ...]
    msg: str
    type: str


class ValidationErrorLogger:
    def __init__(self) -> None:
        self.error_counts: Counter[ValidationError] = Counter()
        self._lock = threading.Lock()

    def log_validation_errors(
        self, consumer: Optional[str], method: str, path: str, detail: List[Dict[str, Any]]
    ) -> None:
        with self._lock:
            for error in detail:
                try:
                    validation_error = ValidationError(
                        consumer=consumer,
                        method=method.upper(),
                        path=path,
                        loc=tuple(str(loc) for loc in error["loc"]),
                        type=error["type"],
                        msg=error["msg"],
                    )
                    self.error_counts[validation_error] += 1
                except (KeyError, TypeError):  # pragma: no cover
                    pass

    def get_and_reset_validation_errors(self) -> List[Dict[str, Any]]:
        data: List[Dict[str, Any]] = []
        with self._lock:
            for validation_error, count in self.error_counts.items():
                data.append(
                    {
                        "consumer": validation_error.consumer,
                        "method": validation_error.method,
                        "path": validation_error.path,
                        "loc": validation_error.loc,
                        "msg": validation_error.msg,
                        "type": validation_error.type,
                        "error_count": count,
                    }
                )
            self.error_counts.clear()
        return data


@dataclass(frozen=True)
class KeyInfo:
    key_id: int
    api_key_id: int
    name: str = ""
    scopes: List[str] = field(default_factory=list)
    expires_at: Optional[datetime] = None

    @property
    def is_expired(self) -> bool:
        return self.expires_at is not None and self.expires_at < datetime.now()

    def has_scopes(self, scopes: Union[List[str], str]) -> bool:
        if isinstance(scopes, str):
            scopes = [scopes]
        if not isinstance(scopes, list):
            raise ValueError("scopes must be a string or a list of strings")
        return all(scope in self.scopes for scope in scopes)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> KeyInfo:
        return cls(
            key_id=data["key_id"],
            api_key_id=data["api_key_id"],
            name=data.get("name", ""),
            scopes=data.get("scopes", []),
            expires_at=(
                datetime.now() + timedelta(seconds=data["expires_in_seconds"])
                if data["expires_in_seconds"] is not None
                else None
            ),
        )


class KeyRegistry:
    def __init__(self) -> None:
        self.salt: Optional[str] = None
        self.keys: Dict[str, KeyInfo] = {}
        self.usage_counts: Counter[int] = Counter()
        self._lock = threading.Lock()

    def get(self, api_key: str) -> Optional[KeyInfo]:
        hash = self.hash_api_key(api_key.strip())
        with self._lock:
            key = self.keys.get(hash)
            if key is None or key.is_expired:
                return None
            self.usage_counts[key.api_key_id] += 1
        return key

    def hash_api_key(self, api_key: str) -> str:
        if self.salt is None:
            raise RuntimeError("Apitally keys not initialized")
        return scrypt(api_key.encode(), salt=bytes.fromhex(self.salt), n=256, r=4, p=1, dklen=32).hex()

    def update(self, keys: Dict[str, Dict[str, Any]]) -> None:
        with self._lock:
            self.keys = {hash: KeyInfo.from_dict(data) for hash, data in keys.items()}

    def get_and_reset_usage_counts(self) -> Dict[int, int]:
        with self._lock:
            data = dict(self.usage_counts)
            self.usage_counts.clear()
        return data



from typing import Optional

from fastapi.exceptions import HTTPException
from fastapi.openapi.models import APIKey, APIKeyIn
from fastapi.requests import Request
from fastapi.security import SecurityScopes
from fastapi.security.base import SecurityBase
from fastapi.security.utils import get_authorization_scheme_param
from starlette.status import HTTP_401_UNAUTHORIZED, HTTP_403_FORBIDDEN

from apitally.client.asyncio import ApitallyClient
from apitally.client.base import KeyInfo
from apitally.starlette import ApitallyMiddleware


__all__ = ["ApitallyMiddleware", "APIKeyAuth", "KeyInfo", "api_key_auth"]


class APIKeyAuth(SecurityBase):
    def __init__(self, *, custom_header: Optional[str] = None, auto_error: bool = True) -> None:
        self.model: APIKey
        if custom_header is None:
            self.model = APIKey(
                **{"in": APIKeyIn.header},  # type: ignore[arg-type]
                name="Authorization",
                description="Provide your API key using the <code>Authorization</code> header and the scheme prefix <code>ApiKey</code>.<br>Example: <pre>Authorization: ApiKey your_api_key_here</pre>",
            )
            self.scheme_name = "Authorization header with ApiKey scheme"
        else:
            self.model = APIKey(
                **{"in": APIKeyIn.header},  # type: ignore[arg-type]
                name=custom_header,
                description=f"Provide your API key using the <code>{custom_header}</code> header.<br>Example: <pre>{custom_header}: your_api_key_here</pre>",
            )
            self.scheme_name = f"{custom_header} header"
        self.auto_error = auto_error

    async def __call__(self, request: Request, security_scopes: SecurityScopes) -> Optional[KeyInfo]:
        header = request.headers.get(self.model.name)
        if self.model.name == "Authorization":
            scheme, api_key = get_authorization_scheme_param(header)
            if not header or scheme.lower() != "apikey":
                if self.auto_error:
                    raise HTTPException(
                        status_code=HTTP_401_UNAUTHORIZED,
                        detail="Not authenticated" if not header else "Invalid authorization scheme",
                        headers={"WWW-Authenticate": "ApiKey"},
                    )
                else:
                    return None  # pragma: no cover
        elif not header:
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_403_FORBIDDEN,
                    detail="Missing API key",
                )
            else:
                return None  # pragma: no cover
        else:
            api_key = header
        key_info = ApitallyClient.get_instance().key_registry.get(api_key)
        if key_info is None and self.auto_error:
            raise HTTPException(
                status_code=HTTP_403_FORBIDDEN,
                detail="Invalid API key",
            )
        if key_info is not None and self.auto_error and not key_info.has_scopes(security_scopes.scopes):
            raise HTTPException(
                status_code=HTTP_403_FORBIDDEN,
                detail="Permission denied",
            )
        if key_info is not None:
            request.state.key_info = key_info
        return key_info


api_key_auth = APIKeyAuth()



from __future__ import annotations

import logging
import sys
import time
from threading import Event, Thread
from typing import Any, Callable, Dict, Optional, Type

import backoff
import requests

from apitally.client.base import (
    ApitallyClientBase,
    ApitallyKeyCacheBase,
    handle_retry_giveup,
)


logger = logging.getLogger(__name__)
retry = backoff.on_exception(
    backoff.expo,
    requests.RequestException,
    max_tries=3,
    on_giveup=handle_retry_giveup,
    raise_on_giveup=False,
)


# Function to register an on-exit callback for both Python and IPython runtimes
try:

    def register_exit(func: Callable[..., Any], *args, **kwargs) -> Callable[..., Any]:  # pragma: no cover
        def callback():
            func()
            ipython.events.unregister("post_execute", callback)

        ipython.events.register("post_execute", callback)
        return func

    ipython = get_ipython()  # type: ignore
except NameError:
    from atexit import register as register_exit


class ApitallyClient(ApitallyClientBase):
    def __init__(
        self,
        client_id: str,
        env: str,
        sync_api_keys: bool = False,
        sync_interval: float = 60,
        key_cache_class: Optional[Type[ApitallyKeyCacheBase]] = None,
    ) -> None:
        super().__init__(
            client_id=client_id,
            env=env,
            sync_api_keys=sync_api_keys,
            sync_interval=sync_interval,
            key_cache_class=key_cache_class,
        )
        self._thread: Optional[Thread] = None
        self._stop_sync_loop = Event()

    def start_sync_loop(self) -> None:
        self._stop_sync_loop.clear()
        if self._thread is None or not self._thread.is_alive():
            self._thread = Thread(target=self._run_sync_loop)
            self._thread.start()
            register_exit(self.stop_sync_loop)

    def _run_sync_loop(self) -> None:
        if self.sync_api_keys:
            with requests.Session() as session:
                self.get_keys(session)
        while not self._stop_sync_loop.is_set():
            try:
                time.sleep(self.sync_interval)
                with requests.Session() as session:
                    self.send_requests_data(session)
                    if self.sync_api_keys:
                        self.get_keys(session)
            except Exception as e:  # pragma: no cover
                logger.exception(e)

    def stop_sync_loop(self) -> None:
        self._stop_sync_loop.set()
        if self._thread is not None:
            self._thread.join()
            self._thread = None

    def send_app_info(self, app_info: Dict[str, Any]) -> None:
        payload = self.get_info_payload(app_info)
        self._send_app_info(payload=payload)

    def send_requests_data(self, session: requests.Session) -> None:
        payload = self.get_requests_payload()
        self._send_requests_data(session, payload)

    def get_keys(self, session: requests.Session) -> None:
        if response_data := self._get_keys(session):  # Response data can be None if backoff gives up
            self.handle_keys_response(response_data)
        elif self.key_registry.salt is None:
            logger.error("Initial Apitally API key sync failed")
            # Exit because the application will not be able to authenticate requests
            sys.exit(1)

    @retry
    def _send_app_info(self, payload: Dict[str, Any]) -> None:
        response = requests.post(url=f"{self.hub_url}/info", json=payload)
        if response.status_code == 404 and "Client ID" in response.text:
            self.stop_sync_loop()
            logger.error(f"Invalid Apitally client ID {self.client_id}")
        else:
            response.raise_for_status()

    @retry
    def _send_requests_data(self, session: requests.Session, payload: Dict[str, Any]) -> None:
        response = session.post(url=f"{self.hub_url}/requests", json=payload)
        response.raise_for_status()

    @retry
    def _get_keys(self, session: requests.Session) -> Dict[str, Any]:
        response = session.get(url=f"{self.hub_url}/keys")
        response.raise_for_status()
        return response.json()



from __future__ import annotations

from typing import TYPE_CHECKING, List, Type

from django.conf import settings
from rest_framework.permissions import BasePermission

from apitally.client.base import KeyInfo
from apitally.client.threading import ApitallyClient
from apitally.django import ApitallyMiddleware


if TYPE_CHECKING:
    from rest_framework.request import Request
    from rest_framework.views import APIView


__all__ = ["ApitallyMiddleware", "HasAPIKey", "HasAPIKeyWithScopes", "KeyInfo"]


class HasAPIKey(BasePermission):  # type: ignore[misc]
    required_scopes: List[str] = []

    def has_permission(self, request: Request, view: APIView) -> bool:
        custom_header = getattr(settings, "APITALLY_CUSTOM_API_KEY_HEADER", None)
        header = request.headers.get("Authorization" if custom_header is None else custom_header)
        if not header:
            return False
        if custom_header is None:
            scheme, _, api_key = header.partition(" ")
            if scheme.lower() != "apikey":
                return False
        else:
            api_key = header
        key_info = ApitallyClient.get_instance().key_registry.get(api_key)
        if key_info is None:
            return False
        if self.required_scopes and not key_info.has_scopes(self.required_scopes):
            return False
        request.auth = key_info
        return True


def HasAPIKeyWithScopes(scopes: List[str]) -> Type[HasAPIKey]:
    class _HasAPIKeyWithScopes(HasAPIKey):  # type: ignore[misc]
        required_scopes = scopes

    return _HasAPIKeyWithScopes
